# (C) Copyright 2019-2021 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from datetime import date

from pyaoscx.api import API


class v10_08(API):
    """
    Represents a REST API Version 10.08. It keeps all the information needed
        for the version and methods related to it.
    """

    def __init__(self):
        self.release_date = date(2021, 6, 21)
        self.version = "10.08"
        self.default_selector = "writable"
        self.default_depth = 1
        self.default_facts_depth = 2
        self.default_subsystem_facts_depth = 4
        self.valid_selectors = [
            "configuration",
            "status",
            "statistics",
            "writable",
        ]
        self.configurable_selectors = ["writable"]
        self.compound_index_separator = ","
        self.valid_depths = [0, 1, 2, 3, 4]

    def _create_ospf_area(self, module_class, session, index_id, **kwargs):
        if "other_config" not in kwargs:
            # If user does not pass value for other_config provide default
            # value, it's needed for correct OSPF Area creation
            kwargs["other_config"] = {
                "stub_default_cost": 1,
                "stub_metric_type": "metric_non_comparable",
            }
        return module_class(session, index_id, **kwargs)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import functools
import json
import logging
import warnings

from abc import ABC, abstractmethod
from copy import deepcopy

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils


class PyaoscxModule(ABC):
    """
    Provide an Interface class for pyaoscx Modules.
    """

    base_uri = ""
    indices = []

    def connected(fnct):
        """
        Function used as a decorator to ensure the module has a established
            connection.
        :param fnct: function which behavior is modified.
        :return ensure_connected: Function.
        """

        @functools.wraps(fnct)
        def ensure_connected(self, *args, **kwargs):
            if not self.session.connected:
                self.session.open()
            return fnct(self, *args, **kwargs)

        return ensure_connected

    def materialized(fnct):
        """
        Function used as a decorator to verify if the object is materialized.
        :para fnct: function which behavior is modified.
        :return is_materialized: Function.
        """

        @functools.wraps(fnct)
        def is_materialized(self, *args, **kwargs):
            if not self.materialized:
                raise VerificationError(
                    "Object {0}".format(self), " not materialized"
                )
            return fnct(self, *args, **kwargs)

        return is_materialized

    def deprecated(func):
        """
        Function used as a decorator to show deprecation notice of a method.
        :param func: function whose behavior is modified/wrapped.
        :return func: function whose behavior is modified/wrapped.
        """

        @functools.wraps(func)
        def is_deprecated(*args, **kwargs):
            warnings.warn(
                "{0} will be removed in a future version".format(
                    func.__name__
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return is_deprecated

    @abstractmethod
    @connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a table entry and fill the
            object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        pass

    @abstractmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system <pyaoscx_module_type> and
            create a dictionary of each object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing object IDs as keys their respective
            objects as values.
        """
        pass

    @abstractmethod
    @connected
    def apply(self):
        """
        Main method used to either create or update an existing
            <pyaoscx_module_type>. Checks whether the <pyaoscx_module_type>
            exists in the switch. Calls self.update() if object being updated.
            Calls self.create() if a new <pyaoscx_module_type> is being
            created.
        :return modified: Boolean, True if object was created or modified.
        """
        pass

    @abstractmethod
    @connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing
            <pyaoscx_module_type> table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        pass

    @abstractmethod
    @connected
    def create(self):
        """
        Perform a POST call to create a new <pyaoscx_module_type>. Only returns
            if an exception is not raise.
        :return modified: Boolean, True if entry was created.
        """
        pass

    @abstractmethod
    @connected
    def delete(self):
        """
        Perform DELETE call to delete <pyaoscx_module_type> table entry.
        """
        pass

    @abstractmethod
    def from_uri(cls, session, uri):
        """
        Create a <pyaoscx_module_type> object given its URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return index, <pyaoscx_module_type>: tuple containing both the
            <pyaoscx_module_type> object its ID.
        """
        pass

    def _get_and_copy_data(
        self, depth=None, selector=None, unwanted_attrs=None
    ):
        """
        Get data from switch with _get_data, and populate the object with the
            data obtained. Note that this is meant to be a common block in the
            get() method of modules derived from PyaoscxModule, and the
            specifics of each module with the _original_attributes dictionary
            should be done in each module's get() body.
        """
        unwanted_attrs = unwanted_attrs or []
        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        self._original_attributes = deepcopy(data)

        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete unwanted attributes
            utils.set_config_attrs(self, data, "config_attrs", unwanted_attrs)

    def _get_data(self, depth, selector):
        """
        Perform a GET call to retrieve data from a switch.
        :param depth: Integer deciding how many levels into the API JSON
            that references will be retrieved from the switch
        :param selector: Alphanumeric option to select specific information
            to return.
        :return: Retrieved data from the switch.
        """
        depth = depth or self.session.api.default_depth

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be one of {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        try:
            response = self.session.request("GET", self.path, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        return json.loads(response.text)

    def _put_data(self, data):
        """
        Perform a PUT request to the switch.
        :param data: data to send.
        :return: True if the object was modified.
        """
        if data == self._original_attributes:
            return False

        self._send_data(self.path, data, "PUT", "Update")
        # Set new original attributes
        self._original_attributes = data
        # Object was modified
        return True

    def _post_data(self, data):
        """
        Perform a POST request to the switch.
        :param data: data to send.
        """
        self._send_data(self.base_uri, data, "POST", "Adding")
        # Get the data from the created object
        self.get()
        return True

    def _send_data(self, path, data, http_verb, display_verb):
        """
        Perform either PUT or POST operation to the switch.
        :param path: path of the resource for the request. This could
            the base URI if this was called in a create method, or
            the VLAN URI if this was called in an update method.
        :param data: data to send.
        :param hrrp_verb: HTTP operation to perfrom.
        :display_module_name: Module to display in logs.
        :display_verb: verb to display in logs.
        """
        send_data = json.dumps(data, sort_keys=True, indent=4)

        try:
            response = self.session.request(http_verb, path, data=send_data)

        except Exception as e:
            raise ResponseError(http_verb, e)

        if not utils._response_ok(response, http_verb):
            raise GenericOperationError(response.text, response.status_code)

        logging.info(
            "SUCCESS: %s %s table entry succeeded",
            display_verb,
            type(self).__name__,
        )

    @staticmethod
    def _is_replace_required(current, replacement, immutable_parameter_names):
        """
        Compares two PYAOSCX modules to determine if a replace (delete+create)
            is required. To do so, all the immutable parameters are checked; in
            case any of them differs a replace is required. Note that if the
            replacement object has a None parameter then a replace is not
            required because they just get ignored, null parameters are
            meant to be taken as 'keep the current value' by PYAOSCX.
        :param current: Module representing the current switch configuration.
        :param replacement: Another Module (same type) object to compare to
        :param immutable_parameter_names: the names of parameters that cannot
            change once the module has been created.
        :return: True if a replacement is required.
        """
        for param_name in immutable_parameter_names:
            if hasattr(current, param_name) and hasattr(
                replacement, param_name
            ):
                # In this case, a common parameter has a different value in the
                # potential replacement config, so a replacement is required
                old = getattr(current, param_name)
                new = getattr(replacement, param_name)
                if new is not None and old != new:
                    return True
            elif hasattr(replacement, param_name):
                # In this case the replacement has an attribute that the
                # current lacks, so a replacement is required
                if getattr(replacement, param_name) is not None:
                    return True

    def _extract_missing_parameters_from(self, other):
        """
        Extract the missing configuration parameters from another PYAOSCX
            Module, to incorporate them as their own. This is useful when
            replacing (delete+create) Modules. If the Module has to be
            replaced, the parameters that are not specified (locally) should
            remain unchanged (switch), so it is necessary to extract them from
            the switch before performing the replacement.
        :param other: the other module to extract the parameters.
        """
        # Until we are able to read the Schema we need to keep a list of the
        # the names of mutable an immutable parameters. Once we are capable of
        # using the schema, the information can be taken from there.
        all_param_names = (
            self.immutable_parameter_names + self.mutable_parameter_names
        )
        for param_name in all_param_names:
            if hasattr(other, param_name):
                param = getattr(other, param_name)
                if not hasattr(self, param_name):
                    if param is not None:
                        setattr(self, param_name, deepcopy(param))
                else:
                    if getattr(self, param_name) is None:
                        setattr(self, param_name, deepcopy(param))

    @deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)



# (C) Copyright 2019-2021 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.interface import Interface as AbstractInterface


class Interface(AbstractInterface):
    """
    Provide configuration management for Interface for Version 10.08.
    """

    pass



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0


class PyaoscxError(Exception):
    """
    Base class for other PYAOSCX exceptions.
    """

    base_msg = "PYAOSCX ERROR"

    def __init__(self, *args):
        self.message = ", ".join(
            (
                self.base_msg,
                *(str(a) for a in args),
            )
        )

    def __str__(self):
        return repr(self.message)



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.device import Device

from pyaoscx.pyaoscx_module import PyaoscxModule


class Qos(PyaoscxModule):
    """
    Provide configuration management for QoS on AOS-CX devices.
    """

    base_uri = "system/qos"
    resource_uri_name = "qos"

    indices = ["name"]

    def __init__(self, session, name, **kwargs):
        """
        Initialize a Qos object.
        :param session: pyaoscx.Session object used to represent logical
            connection to the device.
        :param name: String representing a user-defined name for a Qos object.
        :param uri: a String with an URI.
        """
        self.session = session
        self.__name = name
        # List used to determine attributes related to the QoS configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False
        # Build path
        self.path = "{0}/{1}".format(self.base_uri, self.name)

    @property
    def name(self):
        """
        Method used to obtain the specific name.
        :return: returns the name of this Qos object.
        """
        # This uses the @property decorator to make self.name read-only
        return self.__name

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a QoS table entry and fill the
            object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        if selector in self.session.api.configurable_selectors:
            utils.set_config_attrs(self, data, "config_attrs")
        # Set original attributes
        self._original_attributes = data

        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session):
        """
        Perfom GET request to retrieve all Schedule Profiles in a switch.
        :param session: pyaoscx.Session object used to represent a logical.
        :return: Dictionary containing all system QoS configurations.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        qos_dict = {}

        data = json.loads(response.text)
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            name, qos = cls.from_uri(session, uri)
            qos_dict[name] = qos

        return qos_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Schedule
            Profile. Checks whether the QoS configuration exists in the switch.
            Calls self.update() if object is being updated. Calls self.create()
            if a new object is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if self.materialized:
            self.__modified = self.update()
        else:
            self.__modified = self.create()
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Schedule Profile.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        qos_data = utils.get_attrs(self, self.config_attrs)
        return self._put_data(qos_data)

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST request to create a new QoS configuration using the
            object's attributes as the request's body. An exception is raised
            if object cannot be created.
        :return modified: Boolean, True if entry was created.
        """
        qos_data = utils.get_attrs(self, self.config_attrs)
        qos_data["name"] = self.name
        return self._post_data(qos_data)

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to delete QoS table entry.
        """
        self._send_data(self.path, None, "DELETE", "delete")
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a Qos object given a response_data related to it.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                "<QoS name>": "/rest/v10.08/system/qos/<QoS name>"
            }
        :return: Qos Object.
        """
        qos_name_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )
        qos_name = qos_name_arr[0]
        return cls(session, qos_name)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Qos object given a QoS URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return name, qos: tuple with the name of a QoS configuration, and a
            corresponding object.
        """
        # fail if uri is not a valid QoS URI
        if cls.base_uri not in uri:
            raise ValueError("ERROR: URI must be a valid QoS URI.")

        # Extract name from uri
        name = uri.split("/")[-1]
        qos = cls(session, name)

        return name, qos

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the this instance's URI.
        :return: Object's URI.
        """
        # Parent class uses self.path internally to store the value of the URI
        return self.path

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method to check it object has been modified.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def __str__(self):
        return "Qos {0}".format(self.name)

    @classmethod
    def set_global_trust_mode(cls, session, trust_mode):
        """
        Sets the global trust mode for the switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param trust_mode: A string to set the global trust mode, which can be
            one of the following: "cos", "dscp", "none", or "default".
            Use the default option to use the switch's default trust mode.
        :return: Returns True if configuration was modified.
        """
        logging.info("Setting %s as global Qos trust mode.", trust_mode)

        # Verify trust mode value
        allowed_trust_modes = ["cos", "dscp", "none", "default"]
        if trust_mode not in allowed_trust_modes:
            raise VerificationError(
                "ERROR: QoS trust mode must be one of: ", allowed_trust_modes
            )

        device = Device(session)
        if not device.materialized:
            device.get()
        # If the incoming value is the same as the current one,
        # there's no need to change it
        if device.qos_config.get("qos_trust") == trust_mode:
            return False

        modified = False
        if trust_mode == "default":
            if "qos_trust" in device.qos_config:
                modified = (
                    device.qos_config["qos_trust"]
                    != device.qos_defaults["qos_trust"]
                )
                del device.qos_config["qos_trust"]
        else:
            device.qos_config["qos_trust"] = trust_mode
            modified = True

        device.apply()
        return modified

    @classmethod
    def set_global_schedule_profile(cls, session, profile):
        """
        Sets the global schedule profile (the name of a Qos object) for the
            switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param profile: Name of a schedule profile (the name of a Qos object)
            to set as the global schedule profile.
        """
        logging.info("Setting global schedule profile to: %s", profile)
        device = Device(session)
        device.get()
        setattr(device, "qos_default", profile)
        return device.apply()



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import logging

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class Vsx(PyaoscxModule):
    """
    Provide configuration management for VSX protocol on AOS-CX devices.
    """

    base_uri = "system/vsx"
    path = "system/vsx"

    def __init__(self, session, **kwargs):
        self.session = session
        # List used to determine attributes related to the VSX configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    @property
    def modified(self):
        return self.__modified

    def __get_vrf(self, name):
        vrf = self.session.api.get_module(self.session, "Vrf", name)
        return vrf.get_info_format()

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a VSX table entry and fill the
            class with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)
        # this is common for all PyaoscxModule derived classes
        self._get_and_copy_data(depth, selector)
        # this is specific for this class
        isl_port = self._original_attributes.get("isl_port")
        keepalive_vrf = self._original_attributes.get("keepalive_vrf")
        update_vrf = self._original_attributes.get("software_update_vrf")
        # If the VSX has a isl_port inside the switch and a new one is not
        # being added
        if isl_port:
            isl_port = self.session.api.get_module(
                self.session, "Interface", next(iter(isl_port))
            )
            setattr(self, "isl_port", isl_port.get_info_format())
        # If the VSX has a keepalive_vrf inside the switch and a new one is
        # not being added
        if keepalive_vrf:
            setattr(
                self,
                "keepalive_vrf",
                self.__get_vrf(next(iter(keepalive_vrf))),
            )
        if update_vrf:
            setattr(
                self,
                "software_update_vrf",
                self.__get_vrf(next(iter(update_vrf))),
            )
        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session):
        """
        Not applicable for VSX.
        """

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Vsx object given a VSX URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return Vsx object.
        """
        return cls(session, uri=uri)

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing VSX
            configuration. Checks whether the VSX configuration exists in the
            switch. Calls self.update() if VSX configuration being updated.
            Calls self.create() if a new VSX configuration is being created.
        :return modified: True if object was created or modified.
        """
        if self.materialized:
            return self.update()
        return self.create()

    def _set_isl_port_for_requests(self):
        """
        Set the correct data for the isl_port attribute to use in POST and PUT
            requests.
        """
        if isinstance(self.isl_port, str):
            # if it's a string, it's an interface name, so here it gets changed
            # to an Interface instance
            self.isl_port = self.session.api.get_module(
                self.session, "Interface", self.isl_port
            )
        if not isinstance(self.isl_port, dict):
            # if it's gotten to this point, it should be a either a dictionary,
            # in which case nothing else is needed, or an Interface instance,
            # in which case, the get_info_format() is called to get the
            # dictionary
            self.isl_port = self.isl_port.get_info_format()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing VSX inside switch.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        put_data = utils.get_attrs(self, self.config_attrs)
        if hasattr(self, "keepalive_vrf") and self.keepalive_vrf:
            put_data["keepalive_vrf"] = self.keepalive_vrf
        if hasattr(self, "software_update_vrf") and self.software_update_vrf:
            put_data["software_update_vrf"] = self.software_update_vrf
        if hasattr(self, "isl_port") and self.isl_port:
            self._set_isl_port_for_requests()
            put_data["isl_port"] = self.isl_port
        self.__modified = self._put_data(put_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new VSX. Only returns if no exception
            is raised.
        return: True if entry was created.
        """
        post_data = utils.get_attrs(self, self.config_attrs)
        if hasattr(self, "keepalive_vrf") and self.keepalive_vrf:
            if isinstance(self.keepalive_vrf, str):
                self.keepalive_vrf = self.__get_vrf(self.keepalive_vrf)
            post_data["keepalive_vrf"] = self.keepalive_vrf
        if hasattr(self, "software_update_vrf") and self.software_update_vrf:
            if isinstance(self.software_update_vrf, str):
                self.software_update_vrf = self.__get_vrf(
                    self.software_update_vrf
                )
            post_data["software_update_vrf"] = self.software_update_vrf
        if hasattr(self, "isl_port") and self.isl_port:
            self._set_isl_port_for_requests()
            post_data["isl_port"] = self.isl_port
        if (
            hasattr(self, "keepalive_peer_ip")
            and hasattr(self, "keepalive_src_ip")
            and self.keepalive_src_ip is not None
            and self.keepalive_src_ip is not None
        ):
            ip_src_subnet = self.keepalive_src_ip.find("/")
            ip_peer_subnet = self.keepalive_peer_ip.find("/")
            keepalive_src_ip = self.keepalive_src_ip
            keepalive_peer_ip = self.keepalive_peer_ip
            if ip_src_subnet >= 0:
                keepalive_src_ip = keepalive_src_ip[0:ip_src_subnet]
            if ip_peer_subnet >= 0:
                keepalive_peer_ip = keepalive_peer_ip[0:ip_peer_subnet]
            post_data["keepalive_peer_ip"] = keepalive_peer_ip
            post_data["keepalive_src_ip"] = keepalive_src_ip
        self.__modified = self._post_data(post_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete VSX configuration.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        utils.delete_attrs(self, self.config_attrs)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific VSX URI.
        return: Object's URI.
        """
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Not applicable for VSX.
        """

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.configuration import Configuration

from pyaoscx.pyaoscx_module import PyaoscxModule


class Dns(PyaoscxModule):
    """
    Provide configuration management for DNS on AOS-CX devices. As it is a
        special module, behaves differently.
    """

    base_uri_vrf = "system/vrf"

    def __init__(
        self,
        session,
        vrf_name,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
        uri=None,
    ):

        self.session = session
        self._uri = uri
        # List used to determine attributes related to the DNS configuration
        self.config_attrs = []
        self.materialized = False
        # Attributes needed for DNS
        self.vrf_name = vrf_name
        self.dns_domain_name = domain_name
        self.dns_domain_list = domain_list
        self.dns_name_servers = domain_servers
        self.dns_host_v4_address_mapping = host_v4_address_mapping
        self.dns_host_v6_address_mapping = host_v6_address_mapping

        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        self.create_attrs = [
            "dns_domain_name",
            "dns_domain_list",
            "dns_name_servers",
            "dns_host_v4_address_mapping",
            "dns_host_v6_address_mapping",
        ]
        # Attribute used to know if object was changed recently
        self.__modified = False
        # VRF attribute where configurable attributes go
        self.__internal_vrf = None
        # Try to create VRF
        self.__internal_vrf = self.session.api.get_module(
            self.session, "Vrf", self.vrf_name
        )
        # Materialize internal VRF
        self.__internal_vrf.get()

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a DNS inside the VRF table
            entry and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        # Get VRF
        vrf_obj = self.session.api.get_module(
            self.session, "Vrf", self.vrf_name
        )

        # Get vrf Object
        vrf_obj.get()

        # Sets internal VRF
        self.__internal_vrf = vrf_obj

        # Set attributes with incoming VRF attributes
        if self.dns_domain_name is None:
            self.dns_domain_name = self.__internal_vrf.dns_domain_name
        if self.dns_domain_list is None:
            self.dns_domain_list = self.__internal_vrf.dns_domain_list
        if self.dns_name_servers is None:
            self.dns_name_servers = self.__internal_vrf.dns_name_servers
        if self.dns_host_v4_address_mapping is None:
            self.dns_host_v4_address_mapping = (
                self.__internal_vrf.dns_host_v4_address_mapping
            )
        if self.dns_host_v6_address_mapping is None:
            self.dns_host_v6_address_mapping = (
                self.__internal_vrf.dns_host_v6_address_mapping
            )

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session):
        """
        Method not required for DNS.
        """
        pass

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a new DNS or update an existing DNS,
            configuring it inside the Vrf object. Checks whether the DNS exists
            in the switch. Calls self.update() if DNS configuration is being
            updated.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        # Apply changes
        modified = self.update()

        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing DNS.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        # Obtain variables
        self.__internal_vrf.dns_domain_name = self.dns_domain_name
        self.__internal_vrf.dns_domain_list = self.dns_domain_list
        self.__internal_vrf.dns_name_servers = self.dns_name_servers
        self.__internal_vrf.dns_host_v4_address_mapping = (
            self.dns_host_v4_address_mapping
        )
        self.__internal_vrf.dns_host_v6_address_mapping = (
            self.dns_host_v6_address_mapping
        )

        # Applies changes inside VRF
        modified = self.__internal_vrf.apply()
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Method not implemented.
        """
        pass

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete DNS.
        """
        # Delete the dns settings inside the VRF
        self.dns_domain_name = None
        self.dns_domain_list = None
        self.dns_name_servers = None
        self.dns_host_v4_address_mapping = None
        self.dns_host_v6_address_mapping = None

        # Make changes
        return self.apply()

    @classmethod
    def from_response(cls, session, response_data):
        """
        Not applicable for DNS.
        """
        pass

    @classmethod
    def from_uri(cls, session, uri):
        """
        Not applicable for DNS.
        """
        pass

    def __str__(self):
        return "DNS object with VRF: '{0}'".format(self.vrf_name)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Not applicable for DNS.
        """
        pass

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Not applicable for DNS.
        """
        pass

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def setup_mgmt_nameservers(self, primary=None, secondary=None):
        """
        Setup primary and secondary name servers on a mgmt interface.
        :param primary: Primary nameservers on mgmt interface, a IPv4 address.
            Example: '10.10.2.10'.
        :param secondary: Secondary nameservers on mgmt interface,
            a IP address. Example: '10.10.2.10'.
        :return modified: Return True if coinfig was modified.
        """
        # Create configuration Object
        config = Configuration()

        # Return if configuration was modified
        return config.setup_mgmt_nameservers_dns(primary, secondary)

    def delete_mgmt_nameservers(self):
        """
        Delete primary and secondary name servers on a mgmt interface.
        :return modified: Return True if coinfig was modified.
        """
        # Create configuration Object
        config = Configuration()

        return config.delete_mgmt_nameservers_dns()

    def setup_dns(
        self,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
    ):
        """
        Setup DNS client configuration within a Vrf object.
        :param domain_name: Domain name used for name resolution by the DNS
            client, if 'dns_domain_list' is not configured.
        :param domain_list: dict of DNS Domain list names to be used for
            address resolution, keyed by the resolution priority order.
            Example:
                {
                    0: "hpe.com"
                    1: "arubanetworks.com"
                }
        :param domain_servers: dict of DNS Name servers to be used for address
            resolution, keyed by the resolution priority order. Example:
                {
                    0: "4.4.4.10"
                    1: "4.4.4.12"
                }
        :param host_v4_address_mapping: dict of static host address
            configurations and the IPv4 address associated with them. Example:
                {
                    "host1": "5.5.44.5"
                    "host2": "2.2.44.2"
                }
        :param host_v6_address_mapping: dict of static host address
            configurations and the IPv6 address associated with them. Example:
                {
                    "host1": "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
                }
        :return modified: Returns True if modified.
        """
        # Update Values

        if domain_name is not None:
            self.dns_domain_name = domain_name

        if domain_list is not None:
            self.dns_domain_list = domain_list

        if domain_servers is not None:
            self.dns_name_servers = domain_servers

        if host_v4_address_mapping is not None:
            self.dns_host_v4_address_mapping = host_v4_address_mapping

        if host_v6_address_mapping is not None:
            self.dns_host_v6_address_mapping = host_v6_address_mapping

        return self.apply()

    def delete_dns(
        self,
        domain_name=None,
        domain_list=None,
        domain_servers=None,
        host_v4_address_mapping=None,
        host_v6_address_mapping=None,
    ):
        """
        Delete DNS client configuration within a Vrf object.
        :param domain_name: If value is not None, it is deleted.
        :param domain_list: If value is not None, it is deleted.
        :param domain_servers: If value is not None, it is deleted.
        :param host_v4_address_mapping: If value is not None, it is deleted.
        :param host_v6_address_mapping: If value is not None, it is deleted.
        :return modified: Returns True if modified.
        """
        # Update Values

        if domain_name is not None:
            self.dns_domain_name = None

        if domain_list is not None:
            self.dns_domain_list = None

        if domain_servers is not None:
            self.dns_name_servers = None

        if host_v4_address_mapping is not None:
            self.dns_host_v4_address_mapping = None

        if host_v6_address_mapping is not None:
            self.dns_host_v6_address_mapping = None

        return self.apply()



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class VrfAddressFamily(PyaoscxModule):
    """
    Provide configuration management for Address Family settings on AOS-CX
        devices.
    """

    indices = ["address_family"]
    resource_uri_name = "vrf_address_families"

    def __init__(
        self, session, address_family, parent_vrf, uri=None, **kwargs
    ):

        self.session = session
        # Assign ID
        self.address_family = address_family
        # Assign parent Vrf object
        self.__set_vrf(parent_vrf)
        self._uri = uri
        # List used to determine attributes related to the VRF Address Family
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_vrf(self, parent_vrf):
        """
        Set parent VRF as an attribute for the VrfAddressFamily object.
        :param parent_vrf: a Vrf object.
        """
        # Set parent_vrf
        self.__parent_vrf = parent_vrf

        # Set URI
        self.base_uri = "{0}/{1}/vrf_address_families".format(
            self.__parent_vrf.base_uri, self.__parent_vrf.name
        )

        # Verify VRF Address Family  doesn't exist already inside VRF
        for vrf_address_family in self.__parent_vrf.address_families:
            if vrf_address_family.address_family == self.address_family:
                # Make list element point to current object
                vrf_address_family = self
            else:
                # Add self to vrf_address_families list in parent_vrf
                self.__parent_vrf.address_families.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a VRF Address Family table
            entry and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.address_family)

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the VrfAddressFamily object is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["address_family"]
            )

        # Set original attributes
        self.__original_attributes = data

        # Remove ID
        if "address_family" in self.__original_attributes:
            self.__original_attributes.pop("address_family")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_vrf):
        """
        Perform a GET call to retrieve all system VRF Address Families inside a
            VRF, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where VRF Address Families are stored.
        :return: Dictionary containing VRF Address Family IDs as keys and a
            VrfAddressFamily objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/vrf_address_families".format(
            parent_vrf.base_uri, parent_vrf.name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        vrf_address_family_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a VrfAddressFamily object and adds it to parent
            # VRF list
            address_family, vrf_address_family = VrfAddressFamily.from_uri(
                session, parent_vrf, uri
            )
            # Load all VRF Address Families data from within the Switch
            vrf_address_family.get()
            vrf_address_family_dict[address_family] = vrf_address_family

        return vrf_address_family_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing
            VrfAddressFamily object. Checks whether the VRF Address Family
            exists in the switch. Calls self.update() if VRF Address Family
            being updated. Calls self.create() if a new VRF Address Family is
            being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_vrf.materialized:
            self.__parent_vrf.apply()

        modified = False

        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing VRF Address Family
            table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        vrf_address_family_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(self.base_uri, self.address_family)

        # Compare dictionaries
        if vrf_address_family_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(vrf_address_family_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = vrf_address_family_data
            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new VRF Address Family table entry.
            Only returns if no exception raised.
        return: True if entry was created
        """
        vrf_address_family_data = utils.get_attrs(self, self.config_attrs)
        vrf_address_family_data["address_family"] = self.address_family

        post_data = json.dumps(vrf_address_family_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete VRF Address Family table entry.
        """
        uri = "{0}/{1}".format(self.base_uri, self.address_family)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from VRF
        for vrf_address_family in self.__parent_vrf.address_families:
            if vrf_address_family.address_family == self.address_family:
                self.__parent_vrf.address_families.remove(vrf_address_family)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vrf, response_data):
        """
        Create a VrfAddressFamily object given a response_data related to the
            VRF Address Family's address_family.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where VrfAddressFamlily object is
            stored.
        :param response_data: The response must be a dictionary of the form:
            {
                addr_fam: <URL>
            }
            with URL: "/rest/v10.04/system/vrfs/vrf_address_families/addr_fam"
        :return: VrfAddressFamily object.
        """
        vrf_address_family_arr = session.api.get_keys(
            response_data, VrfAddressFamily.resource_uri_name
        )
        address_family = vrf_address_family_arr[0]
        return VrfAddressFamily(session, address_family, parent_vrf)

    @classmethod
    def from_uri(cls, session, parent_vrf, uri):
        """
        Create a VrfAddressFamily object given a URI and parent VRF.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent vrf class where VRF Address Family is stored.
        :param uri: a String with a URI.
        :return index, vrf_address_family_obj: tuple containing both the VRF
            Address Family object and the VRF Address Family's address_family.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)vrf_address_families/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")
        # Create VrfAddressFamily object
        vrf_address_family_obj = VrfAddressFamily(
            session, index, parent_vrf, uri=uri
        )

        return index, vrf_address_family_obj

    def __str__(self):
        return "VRF Address Family ID {0}".format(self.address_family)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific VRF Address Family URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.address_family,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



from setuptools import setup
from setuptools import find_packages

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pyaoscx",
    version="2.2.1",
    description="AOS-CX Python Modules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aruba/pyaoscx",
    author="Aruba Automation",
    author_email="aruba-automation@hpe.com",
    license="Apache 2.0",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Topic :: System :: Networking",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="networking aruba aos-cx switch rest api python",
    packages=find_packages(exclude=["docs"]),
    install_requires=["requests", "PyYAML", "netaddr"],
    zip_safe=False,
)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.generic_op_error import GenericOperationError


class ListDescriptor(list):
    """
    Attribute descriptor class to keep track of a list that contains
        pyaoscx_module objects simulating a Reference to a resource. If the
        list changes, then every pyaoscx_module object has to be changed.
    """

    def __init__(
        self,
        name,
    ):
        self.name = name

    def __get__(self, instance, owner):
        """
        Method called when current attribute is used.
        :param instance: Instance of the current Object
        """
        return instance.__dict__[self.name]

    def __set__(self, instance, new_list):
        """
        Method called when current attribute is set.
        :param instance: Instance of the current Object.
        :param new_list: new list being set to current attribute object.
        """
        new_list = ReferenceList(new_list)
        prev_list = (
            instance.__dict__[self.name]
            if self.name in instance.__dict__
            else None
        )

        # Update value inside the instance dictionary
        instance.__dict__[self.name] = new_list

        # Check changes and delete
        if prev_list is not None and prev_list != new_list:

            # Reflect changes made inside the list
            for element in prev_list:
                if element not in new_list:
                    # Delete element reference
                    try:
                        element.delete()
                    except AttributeError:
                        # Ignore
                        pass


class ReferenceList(list):
    """
    Wrapper class for a Python List object.
        Modifies remove() method to use the pyaoscx.pyaoscx_module.delete()
        method when using remove on this special type list.
    """

    def __init__(self, value):
        list.__init__(self, value)

    def __setitem__(self, key, value):
        """
        Intercept the l[key]=value operations.
        Also covers slice assignment.
        """
        try:
            _ = self.__getitem__(key)
        except KeyError:
            list.__setitem__(self, key, value)
        else:
            list.__setitem__(self, key, value)

    def __delitem__(self, key):
        """
        Delete self.key.
        """
        _ = list.__getitem__(self, key)
        list.__delitem__(self, key)

    def pop(self):
        """
        Remove and return item at index (default last).
        """
        oldvalue = list.pop(self)
        return oldvalue

    def extend(self, newvalue):
        """
        Extend list by appending elements from iterable.
        """
        list.extend(self, newvalue)

    def insert(self, i, element):
        """
        Insert object before index.
        """
        list.insert(self, i, element)

    def remove(self, element):
        """
        Remove first occurrence of value.
        """
        _ = list.index(self, element)
        list.remove(self, element)
        try:
            # Delete element with a DELETE request
            element.delete()

        # If delete fails because table entry
        # is already deleted: IGNORE
        except GenericOperationError as error:
            # In case error is not 404, raise
            if error.response_code != 404:
                raise error

    def reverse(self):
        """
        Reverse *IN PLACE*.
        """
        list.reverse(self)

    def sort(self, cmpfunc=None):
        """
        Stable sort *IN PLACE*.
        """
        _ = self[:]
        list.sort(self, cmpfunc)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import getpass
import json
import logging
import re

import requests

from pyaoscx.exceptions.login_error import LoginError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.api import API

# Global Variables
ZEROIZED = 268
UNAUTHORIZED = 401


class Session:
    """
    Represents a connection to an AOS-CX device. It keeps all information
        needed to login/logout to it, including parameters like proxy, IP
        address (both IPv4 and IPv6 are supported), and API version. The IP
        address should be similar to:
        '1.1.1.1'
        or
        '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
    """

    def __init__(self, ip_address, api, proxy=None):

        self.api = API.create(api)
        self.ip = ip_address
        self.connected = False
        self.proxy = (
            {"no_proxy": self.ip} if proxy is None else {"https": proxy}
        )
        self.scheme = "https"
        self.version_path = "rest/v{0}/".format(self.api)

        # TODO: remove base_url once all modules use the internal
        # request methods
        self.base_url = "https://{0}/rest/v{1}/".format(self.ip, self.api)
        self.resource_prefix = "/rest/v{0}/".format(self.api)
        self.s = requests.Session()
        self.s.verify = False
        self.__username = self.__password = ""

    def cookies(self):
        """
        Return the cookie stored in the requests' session.
        """
        return self.s.cookies._cookies

    @classmethod
    def from_session(cls, req_session, base_url, credentials=None):
        """
        Create a session from an existing request Session. It allows to create
            an internal session from an already-authenticated and serialized
            session.
        :param req_session: Existing Request Session object.
        :param base_url: Url needed to create Session Object.
        :param credentials: Dictionary with user and password credentials.
            Example: {
                username: <username>,
                password: <password>
            }
        :return session: Request Session Object.
        """
        ip_address = version = ""

        # From base url retrieve the ip address and the version
        url_pattern = re.compile(
            r"https://(?P<ip_address>.+)/rest/v(?P<version>.+)/"
        )
        match = url_pattern.match(base_url)
        if match:
            ip_address = match.group("ip_address")
            version = match.group("version")

        else:
            raise Exception("Error creating Session")

        # Create Session
        session = Session(ip_address, version)

        # Determine proxy
        # If the proxy is not {} then it would replace the proxy
        # previously created inside the __init__ method
        if req_session.proxies != {}:
            session.proxy = req_session.proxies

        # Set request.Session()
        session.s.cookies = req_session.cookies
        session.connected = True

        # Set credentials
        if credentials is not None:
            session.__username = credentials["username"]
            session.__password = credentials["password"]

        return session

    def open(self, username=None, password=None):
        """
        Perform a POST call to login and gain access to other API calls. If
            either username or password is not specified, user will be prompted
            to enter the missing credential(s).
        :param username: username
        :param password: password
        """
        if self.__username is None:
            if username is None:
                username = input("Enter username: ")
            else:
                self.__username = username
        else:
            if username is None:
                username = input("Enter username: ")

        if self.__password is None:
            if password is None:
                password = getpass.getpass()
            else:
                self.__password = password
        else:
            if password is None:
                password = getpass.getpass()

        login_data = {"username": username, "password": password}
        self.__username = username
        self.__password = password
        try:
            login_uri = "{0}{1}".format(self.base_url, "login")
            response = self.s.post(
                login_uri,
                data=login_data,
                verify=False,
                timeout=5,
                proxies=self.proxy,
            )
        except requests.exceptions.ConnectTimeout:
            raise Exception(
                "Error connecting to host: connection attempt timed out."
            )

        if response.status_code != 200:
            raise Exception(
                "FAIL: Login failed with status code {0}: {1}".format(
                    response.status_code, response.text
                )
            )

        cookies = self.s.cookies
        self.connected = (
            hasattr(cookies, "_cookies") and self.ip in cookies._cookies
        )

        if not self.connected:
            raise LoginError("Cookies were not set correctly. Login failed")

        logging.info("SUCCESS: Login succeeded")

    def close(self):
        """
        Perform a POST call to logout and end the session. Given all the
            required information to perform the operation is already stored
            within the session, no parameters are required.
        """
        if self.connected:
            logout_uri = "{0}{1}".format(self.base_url, "logout")

            try:
                response = self.s.post(
                    logout_uri, verify=False, proxies=self.proxy
                )
            except BaseException:
                raise Exception(
                    "Unable to process the request: ({0}) {1}".format(
                        response.status_code, response.text
                    )
                )

            if response.status_code != 200:
                raise Exception(
                    "FAIL: Logout failed with status code {0}: {1}".format(
                        response.status_code, response.text
                    )
                )

            logging.info("SUCCESS: Logout succeeded")

    # Session Login and Logout
    # Used for Connection within Ansible

    @classmethod
    def login(
        cls,
        base_url,
        username=None,
        password=None,
        use_proxy=True,
        handle_zeroized_device=False,
    ):
        """
        Perform a POST call to login and gain access to other API calls. If
            either username or password is not specified, user will be prompted
            to enter the missing credential(s).
        :param base_url: URL in main() function
        :param username: username
        :param password: password
        :param use_proxy: Whether the system proxy should be used, defaults to
            True.
        :param handle_zeroized_device: Whether a zeroized device should be
            initialized, if so sets the admin password to the provided one,
            defaults to False.
        :return: requests.session object with loaded cookie jar
        """
        if username is None and password is None:
            username = input("Enter username: ")
            password = getpass.getpass()

        login_data = {"username": username, "password": password}

        s = requests.Session()

        if use_proxy is False:
            s.proxies["https"] = None
            s.proxies["http"] = None
        try:
            print(base_url + "login")
            response = s.post(
                base_url + "login",
                data=login_data,
                verify=False,
                timeout=5,
                proxies=s.proxies,
            )
        except requests.exceptions.ConnectTimeout:
            logging.warning(
                "ERROR: Error connecting to host: "
                "connection attempt timed out."
            )
            raise LoginError(
                "ERROR: Error connecting to host: "
                "connection attempt timed out."
            )
        except requests.exceptions.ProxyError as err:
            logging.warning("ERROR: %s", str(err))
            raise LoginError("ERROR: {0}".format(err))
        # Response OK check needs to be passed 'PUT' since this
        # POST call returns 200 instead of conventional 201
        if not utils._response_ok(response, "PUT"):
            if response.status_code == UNAUTHORIZED and handle_zeroized_device:
                # Try to login with default credentials:
                ztp_login_data = {"username": username}
                response = s.post(
                    base_url + "login",
                    data=ztp_login_data,
                    verify=False,
                    timeout=5,
                    proxies=s.proxies,
                )
                if response.status_code == ZEROIZED:
                    data = {"password": password}
                    response = s.put(
                        base_url + "system/users/admin",
                        data=json.dumps(data),
                        verify=False,
                        timeout=5,
                        proxies=s.proxies,
                    )
                    if utils._response_ok(response, "PUT"):
                        logging.info("SUCCESS: Login succeeded")
                        return s
            logging.warning(
                "FAIL: Login failed with status code %d: %s",
                response.status_code,
                response.text,
            )
            raise LoginError(
                "FAIL: Login failed with status code {0}: {1}".format(
                    response.status_code, response.text
                ),
                response.status_code,
            )
        else:
            logging.info("SUCCESS: Login succeeded")
            return s

    @classmethod
    def logout(cls, **kwargs):
        """
        Perform a POST call to logout and end session.
        :param kwargs:
            keyword s: requests.session object with loaded cookie jar
            keyword url: URL in main() function
        :return: True if successful.
        """
        response = kwargs["s"].post(
            kwargs["url"] + "logout", verify=False, proxies=kwargs["s"].proxies
        )
        # Response OK check needs to be passed 'PUT' since this POST
        # call returns 200 instead of conventional 201
        if not utils._response_ok(response, "PUT"):
            logging.warning(
                "FAIL: Logout failed with status code %d: %s",
                response.status_code,
                response.text,
            )
            return False
        else:
            logging.info("SUCCESS: Logout succeeded")
            return True

    def username(self):
        """
        Get username.
        :return username.
        """
        return self.__username

    def password(self):
        """
        Get password.
        :return password.
        """
        return self.__password

    def _build_uri(self, resource_path):
        """
        Build a URI representing a resource.
        :param resource_path: Resource path before adding version prefix.
        :return: String of the uri
        """
        complete_path = self.version_path + resource_path
        uri = requests.utils.urlunparse(
            (self.scheme, self.ip, complete_path, "", "", "")
        )
        return uri

    def request(self, operation, path, params=None, data=None, verify=False):
        """
        Perform a Request to the switch.
        :param operation: type of operation: PUT, GET, POST, DELETE.
        :param path: Path to the resource.
        :param params: Extra request parameters.
        :param data: Data to send in the resquest.
        :param verify: If session should verify.
        :return: response object from the request.
        """
        operations = {
            "PUT": self.s.put,
            "GET": self.s.get,
            "POST": self.s.post,
            "DELETE": self.s.delete,
        }

        if operation not in operations:
            raise VerificationError(
                "The operation {0} is not supported."
                " Use any of {1}".format(operation, list(operations.keys))
            )

        return operations[operation](
            self._build_uri(path),
            verify=verify,
            params=params,
            data=data,
            proxies=self.proxy,
        )



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.pyaoscx_error import PyaoscxError


class GenericOperationError(PyaoscxError):
    """
    PYAOSCX Generic Operation Error Exception.
    """
    base_msg = "GENERIC OPERATION ERROR"

    def __init__(self, *args):
        self.message = None
        self.response_code = None
        self.extra_info = None
        if args:
            self.message = args[0]
            if len(args) >= 2:
                self.response_code = args[1]
            if len(args) > 2:
                self.extra_info = ", ".join(str(a) for a in args[2:])

    def __str__(self):
        msg_parts = [self.base_msg]
        if self.message:
            msg_parts.append(str(self.message))
        if self.response_code:
            msg_parts.append("Code")
            msg_parts.append(str(self.response_code))
        if self.extra_info:
            msg_parts.append("on Module")
            msg_parts.append(str(self.extra_info))
        msg = ": ".join(msg_parts)
        return repr(msg)



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.unsupported_capability_error import (
    UnsupportedCapabilityError,
)
from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule

from pyaoscx.device import Device
from pyaoscx.qos import Qos


class Queue(PyaoscxModule):
    """
    Provide configuration management for Queues on AOS-CX devices.
    """

    indices = ["queue_number", "qos_name"]
    resource_uri_name = "queues"

    def __init__(self, session, qos_name, queue_number, **kwargs):
        """
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param qos_name: String with a user-defined name for a Qos object.
        :param queue_number: Integer representing a queue priority, which are
            numbered in priority order, with zero being the lowest priority.
            The maximum number of queues is hardware dependent.
        """
        self.session = session
        self.__queue_number = queue_number
        self.__qos_name = qos_name
        # this is needed for the property, this is safe
        self.__gmb_percent = None
        self.__burst = None
        # List of configuration attributes
        self.config_attrs = []

        # Get and remove burst from kwargs if given
        if "burst" in kwargs:
            self.burst = kwargs.pop("burst")

        # Get and remove gmb_percent from kwargs if given
        if "gmb_percent" in kwargs:
            self.gmb_percent = kwargs.pop("gmb_percent")

        # Original attributes
        self._original_attributes = {}
        self.materialized = False
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

        self.base_uri = "{0}/{1}/{2}".format(
            Qos.base_uri, qos_name, self.resource_uri_name
        )

        self.path = "{0}/{1}".format(self.base_uri, self.queue_number)

    @property
    def burst(self):
        return self.__burst

    @burst.setter
    def burst(self, value):
        if not Device(self.session).is_capable("qos_queue_burst"):
            raise UnsupportedCapabilityError(
                "This device can't configure a Queue's burst."
            )
        self.__burst = value

    @property
    def gmb_percent(self):
        return self.__gmb_percent

    @gmb_percent.setter
    def gmb_percent(self, value):
        if not Device(self.session).is_capable("qos_sched_min_bandwidth"):
            raise UnsupportedCapabilityError(
                "This device can't configure a Queue's minimum bandwidth."
            )
        self.__gmb_percent = value

    @property
    def queue_number(self):
        """
        Method to retrieve the queue_number identifier of this object.
        :return: returns the queue number of this object.
        """
        return self.__queue_number

    @property
    def qos_name(self):
        """
        Method to retrieve the qos_name identifier of this object.
        :return: returns the Qos name of this object.
        """
        return self.__qos_name

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Queue table entry and fill
            object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        for k, v in data.items():
            setattr(self, k, v)

        if selector in self.session.api.configurable_selectors:
            self.config_attrs = list(data)

        # Set original attributes
        self._original_attributes = data

        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session, qos_name):
        """
        Perform a GET call to retrieve all system Queues for given Schedule
            Profile from a switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing all system Schedule Profile's Queues.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}{1}/{2}/{3}".format(
            cls.base_uri, Qos.base_uri, qos_name, cls.resource_uri_name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        queues_dict = {}

        data = json.loads(response.text)
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            number, queue = cls.from_uri(session, uri)
            queues_dict[number] = queue

        return queues_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Queue. Checks
            whether the Queue exists in the switch. Calls self.update() if
            object is being updated.
        Calls self.create() if a new object is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if self.materialized:
            self.__modified = self.update()
        else:
            self.__modified = self.create()
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Queue table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        queue_data = utils.get_attrs(self, self.config_attrs)
        return self._put_data(queue_data)

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST request to create a new Queue using the object's
            attributes as the request body. An exception is raised if object
            cannot be created.
        :return modified: Boolean, True if entry was created.
        """
        queue_data = utils.get_attrs(self, self.config_attrs)
        queue_data["queue_number"] = self.queue_number

        return self._post_data(queue_data)

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to delete Queue table entry.
        """
        self._send_data(self.path, None, "DELETE", "delete")

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a Queue object given a related response_data.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                "strict": "/rest/v10.08/system/qos/"<Qos name>/queues/7"
            }
        :return: Queue Object.
        """
        # Check if response is a dictionary, if so, get its value
        if isinstance(response_data, dict):
            data = list(response_data.items())[0][1]
        else:
            # when not a dictionary, it's a string
            data = response_data

        # Get queue number from uri
        data_arr = data.split("/")
        queue_number = data_arr[-1]
        qos_name = data_arr[data_arr.index(Qos.resource_uri_name) + 1]

        return cls(session, qos_name, queue_number)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Queue object given a Queue URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return name, queue: tuple with the Queue object and its name.
        """
        # Get queue number from uri
        uri_arr = uri.split("/")
        queue_number = uri_arr[-1]
        qos_name = uri_arr[uri_arr.index(Qos.resource_uri_name) + 1]
        queue = cls(session, qos_name, queue_number)

        return queue_number, queue

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain this instance's URI.
        :return: Object's URI.
        """
        return self.path

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method to check it object has been modified.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def __str__(self):
        return "Queue {0}".format(self.queue_number)



# (C) Copyright 2020-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import logging

from pyaoscx.utils import util as utils


def get_firmware_version(**kwargs):
    """
    Perform GET calls to retrieve the current firmware version.
    :param kwargs:
        keyword s: requests.session object with loaded cookie jar
        keyword url: URL in main() function.
    :return: Firmware version string if found, otherwise None.
    """
    target_url = kwargs["url"] + "firmware"

    response = kwargs["s"].get(
        target_url, verify=False, proxies=kwargs["s"].proxies
    )

    if not utils._response_ok(response, "GET"):
        logging.warning(
            "FAIL: Getting firmware version %d: %s",
            response.status_code,
            response.text,
        )
        firmware_version = None
    else:
        logging.info("SUCCESS: Getting firmware version succeeded")
        firmware_version = response.json()["current_version"]

    return firmware_version



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from netaddr import IPNetwork

import requests

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.parameter_error import ParameterError


def create_attrs(obj, data_dictionary):
    """
    Given a dictionary object creates class attributes. The methods implements
        setattr() which sets the value of the specified attribute of the
        specified object. If the attribute is already created within the
        object. It's state changes only if the current value is not None.
        Otherwise it keeps the previous value.
    :param data_dictionary: dictionary containing the attributes.
    """
    import copy

    # Used to create a deep copy of the dictionary
    dictionary_var = copy.deepcopy(data_dictionary)

    # K is the argument and V is the value of the given argument
    for k, v in dictionary_var.items():
        # In case a key has '-' inside it's name.
        k = k.replace("-", "_")
        obj.__dict__[k] = v


def get_dict_keys(dict):
    """
    Function used to get a list of all the keys of the respective dictionary.
    :param dict: Dictionary object used to obtain the keys.
    :return: List containing the keys of the given dictionary.
    """
    list = []
    for key in dict.keys():
        list.append(key)

    return list


def check_args(obj, **kwargs):
    """
    Given a object determines if the coming arguments are not already inside
        the object. If attribute is inside the config_attrs, it is ignored.
    :param obj: object in which the attributes are being set to.
    :param **kwargs list of arguments used to create the attributes.
    :return correct: True if all arguments are correct.
    """
    arguments = get_dict_keys(kwargs)
    correct = True
    for argument in arguments:
        if hasattr(obj, argument):
            correct = False
    return correct


def delete_attrs(obj, attr_list):
    """
    Given an object and a list of strings, delete attributes with the same name
        as the one inside the list.
    :param attr_list: List of attribute names that will be deleted from object
    """
    for attr in attr_list:
        if hasattr(obj, attr):
            delattr(obj, attr)


def get_attrs(obj, config_attrs):
    """
    Given an object obtains the attributes different to None.
    :param obj: object containing the attributes.
    :param config_attrs: a list of all the configurable attributes within the
        object.
    :return attr_data_dict: A dictionary containing all the attributes of the
        given object that have a value different to None.
    """
    attr_data_dict = {}
    for attr_name in config_attrs:
        attr_data_dict[attr_name] = getattr(obj, attr_name)
    return attr_data_dict


def set_creation_attrs(obj, **kwargs):
    """
    Used when instantiating the class with new attributes. Sets the
        configuration attributes list, for proper management of attributes
        related to configuration.
    :param obj: Python object in which attributes are being set.
    :param **kwargs: a dictionary containing the possible future arguments for
        the object.
    """
    if check_args(obj, **kwargs):
        obj.__dict__.update(kwargs)
        set_config_attrs(obj, kwargs)
    else:
        raise Exception(
            "ERROR: Trying to create existing attributes inside the object"
        )


def set_config_attrs(
    obj, config_dict, config_attrs="config_attrs", unwanted_attrs=[]
):
    """
    Add a list of strings inside the object to represent each attribute for
        config purposes.
    :param config_dict: Dictionary where each key represents an attribute.
    :param config_attrs: String containing the name of the attribute referring
        to a list.
    :param unwanted_attrs: Attributes that should be deleted, since they can't
        be modified.
    """
    # Set new configuration attributes list
    new_config_attrs = get_dict_keys(config_dict)

    # Delete unwanted attributes from configuration attributes list
    for element in unwanted_attrs:
        if element in new_config_attrs:
            # Remove all occurrences of element inside
            # the list representing the attributes related
            # to configuration
            new_config_attrs = list(filter((element).__ne__, new_config_attrs))
    # Set config attributes list with new values
    obj.__setattr__(config_attrs, new_config_attrs)


def _response_ok(response, call_type):
    """
    Checks whether API HTTP response contains the associated OK code.
    :param response: Response object.
    :param call_type: String containing the HTTP request type.
    :return: True if response was OK.
    """
    ok_codes = {
        "GET": [200],
        "PUT": [200, 204],
        "POST": [201],
        "DELETE": [204],
    }

    return response.status_code in ok_codes[call_type]


def file_upload(session, file_path, complete_uri):
    """
    Upload any file given a URI and the path to a file located on the local
        machine.
    :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
    :param file_path: File name and path for local file uploading.
    :param complete_uri: Complete URI to perform the POST Request and upload
        the file. Example:
            "https://172.25.0.2/rest/v10.04/firmware?image=primary".
    :return True if successful.
    """
    with open(file_path, "rb") as file:
        file_param = {"fileupload": file}
        try:
            # User session login
            # Perform Login
            response_login = requests.post(
                "{0}login?username={1}&password={2}".format(
                    session.base_url, session.username(), session.password()
                ),
                verify=False,
                timeout=5,
                proxies=session.proxy,
            )

            # Perform File Upload
            response_file_upload = requests.post(
                url=complete_uri,
                files=file_param,
                verify=False,
                proxies=session.proxy,
                cookies=response_login.cookies,
            )

            # Perform Logout
            requests.post(
                session.base_url + "logout",
                verify=False,
                proxies=session.proxy,
                cookies=response_login.cookies,
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if response_file_upload.status_code != 200:
            raise GenericOperationError(
                response_file_upload.text, response_file_upload.status_code
            )

    # Return true if successful
    return True


def get_ip_version(ip):
    """
    Map if given IP address is v4 or v6. Will raise an exception if given
        address if invalid.
    :param ip: String with an IP address.
    :return: String with the IP version. Can be either ipv4 or ipv6.
    """
    try:
        ip_net = IPNetwork(ip)
        return "ipv{0}".format(ip_net.version)
    except ValueError:
        msg = "Invalid IP Address: {0}".format(ip)
        raise ParameterError(msg)



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.parameter_error import ParameterError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class OspfVlink(PyaoscxModule):
    """
    Provide configuration management for OSPF VLink instance on AOS-CX devices.
    """

    collection_uri = (
        "system/vrfs/{name}/ospf{version}_routers/"
        "{instance_tag}/areas/{area_id}/ospf_vlinks"
    )
    object_uri = collection_uri + "/{peer_router_id}"

    resource_uri_name = "ospf_vlinks"

    def __init__(self, session, peer_router_id, parent_ospf_area, **kwargs):
        self.session = session
        self.__parent_ospf_area = parent_ospf_area
        self.__peer_router_id = peer_router_id
        # List used to determine attributes related to the OPSF configuration
        self.config_attrs = []
        self.materialized = False
        # Dictionary used to manage original data obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False
        self.base_uri = self.__parent_ospf_area.path + "/ospf_vlinks"
        self.path = "{0}/{1}".format(self.base_uri, self.__peer_router_id)

    @property
    def peer_router_id(self):
        """
        Return this object's identifier.
        """
        return self.__peer_router_id

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    def _get_indices(self):
        """
        Get indices to retrieve collection of this object's instances.
        :return: a dictionary with each key in the collection_uri, and its
            respective value to perform a GET request, or empty dictionary
            if the collection_uri has no indices.
        """
        indices = {"area_id": self.__parent_ospf_area.area_id}
        indices.update(self.__parent_ospf_area._get_indices())
        return indices

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET request to retrieve data for an OSPF VLink table entry
            and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Set original attributes
        self._original_attributes = data

        # Sets object as materialized
        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session, parent_ospf_area):
        """
        Perform a GET request to retrieve all system OSPF Virtual Links inside
            a OPSF Router, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_area: parent OPSF Area object where OPSF VLink
            is stored.
        :return: Dictionary containing OSPF Virtual Link IDs as keys and a OSPF
            Virtual Link objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri_indices = {"area_id": parent_ospf_area.area_id}
        uri_indices.update(parent_ospf_area._get_indices())
        uri = cls.collection_uri.format(uri_indices)
        try:
            response = session.request("GET", uri)
        except Exception as exc:
            raise ResponseError("GET", exc) from exc
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)
        data = json.loads(response.text)
        ospf_vlink_dict = {}
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            peer_router_id, vlink = cls.from_uri(
                session, uri, parent_ospf_area
            )
            vlink.get()
            ospf_vlink_dict[peer_router_id] = vlink

        return ospf_vlink_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Ospf Virtual
            Link. Checks whether the OSPF Virtual Link exists in the switch.
            Calls self.update() if OSPF Virtual Link is being updated. Calls
            self.create() if a new OSPF Virtual Link is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_ospf_area.materialized:
            self.__parent_ospf_area.apply()

        if self.materialized:
            self.__modified = self.update()
        else:
            self.__modified = self.create()
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT request to apply changes to an existing OSPF VLink.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        ospf_vlink_data = utils.get_attrs(self, self.config_attrs)
        self.__modified = self._put_data(ospf_vlink_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST request to create a new OSPF Virtual Link. Only returns
            if an exception is not raised.
        :return modified: Boolean, True if object was created.
        """
        ospf_vlink_data = utils.get_attrs(self, self.config_attrs)
        ospf_vlink_data["peer_router_id"] = self.__peer_router_id

        self.__modified = self._post_data(ospf_vlink_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete OSPF Virtual Link table entry.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_uri(cls, session, uri, parent_ospf_area=None):
        """
        Create an OspfVlink object given a URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_area: parent OspfArea object where OspfVlink object
            is stored.
        :param uri: an OSPF VLink URI with its index (a peer_router_id).
        :return peer_router_id, ospf_vlink: tuple with the OspfVlink ID, and
            the object.
        """
        if parent_ospf_area is None:
            raise ParameterError(
                "{0} requires parent_ospf_area instance".format(cls.__name__)
            )
        # Obtain ID from URI of the form '.../ospf_vlinks/{peer_router_id}'
        peer_router_id = uri.split("/")[-1]
        vlink = cls(session, peer_router_id, parent_ospf_area)
        return peer_router_id, vlink

    def __str__(self):
        return "OSPF Virtual Link ID {0}".format(self.__peer_router_id)

    def get_uri(self):
        """
        Method used to obtain the specific OSPF Virtual Link URI.
        return: Object's URI.
        """
        # PyaoscxModule uses the uri with the name self.path
        # so it needs to have that name
        return self.path

    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified
